/*********************************************************************/
/*        F u n c t i o n     P R A X I S                            */
/*                                                                   */
/* PRAXIS is a general purpose routine for the minimization of a     */
/* function in several variables. the algorithm used is a modifi-    */
/* cation of conjugate gradient search method by powell. the changes */
/* are due to r.p. brent, who gives an algol-w program, which served */
/* as a basis for this function.                                     */
/*                                                                   */
/* references:                                                       */
/*     - powell, m.j.d., 1964. an efficient method for finding       */
/*       the minimum of a function in several variables without      */
/*       calculating derivatives, computer journal, 7, 155-162       */
/*     - brent, r.p., 1973. algorithms for minimization without      */
/*       derivatives, prentice hall, englewood cliffs.               */
/*                                                                   */
/*     problems, suggestions or improvements are always wellcome     */
/*                       karl gegenfurtner   07/08/87                */
/*                                           c - version             */
/*********************************************************************/
/*                                                                   */
/* usage: min = praxis(fun, x, h0, n);                               */
/*                                                                   */
/*   fun       the function to be minimized. fun is called from      */
/*             PRAXIS with x and n as arguments                      */
/*   x         a double array containing the initial guesses for     */
/*             the minimum, which will contain the solution on       */
/*             return                                                */
/*   h0        is the maximum step size: Set to about the maximum    */
/*             expected distance from the guess to the minimum       */
/*             (if H is set too small or too large then the initial  */
/*             rate of convergence will be slow).                    */
/*   n         an integer specifying the number of unknown           */
/*             parameters                                            */
/*   min       PRAXIS returns the least calculated value of fun      */
/*                                                                   */
/* some additional global variables control some more aspects of     */
/* the inner workings of PRAXIS. setting them is optional, they      */
/* are all set to some reasonable default values given below.        */
/*                                                                   */
/*   prin      controls the printed output from the routine.         */
/*             0 -> no output                                        */
/*             1 -> print only starting and final values             */
/*             2 -> detailed map of the minimization process         */
/*             3 -> print also eigenvalues and vectors of the        */
/*                  search directions                                */
/*             the default value is 1                                */
/*   tol       is the tolerance allowed for the precision of the     */
/*             solution. PRAXIS returns if the criterion             */
/*             2 * ||x[k]-x[k-1]|| <= sqrt(macheps) * ||x[k]|| + tol */
/*             is fulfilled more than ktm times.                     */
/*             the default value depends on the machine precision    */
/*   ktm       see just above. default is 1, and a value of 4 leads  */
/*             to a very(!) cautious stopping criterion.             */
/*   scbd      is a scaling parameter. 1.0 is the default and        */
/*             indicates no scaling. if the scales for the different */
/*             parameters are very different, scbd should be set to  */
/*             a value of about 10.0.                                */
/*   illc      should be set to true (1) if the problem is known to  */
/*             be ill-conditioned. the default is false (0). this    */
/*             variable is automatically set, when PRAXIS finds      */
/*             the problem to be ill-conditioned during iterations.  */
/*                                                                   */
/*********************************************************************/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "machine.h"

/* void minfit(int, double, double, double (*)[N], double *); */
void minfit(int, double, double, double [N][N], double [N]);
void diagramm(int, double, char *);
void report(int, int, double, int, int);

double RandBR();

/* control parameters */
double tol = 0.00001,
       scbd = 1.0;
int    ktm = 1,
       illc = 0,
       prin = 1;

/* some global variables */
static int i, j, k, k2, nl, nf, kl, kt;
static double s, sl, dn, dmin,
       fx, f1, lds, ldt, sf, df,
       m2, m4, small, vsmall, large,
       vlarge, ldfac, t2;
static double d[N], y[N], z[N], v[N][N], tflin[N];

/* Quad - Variablen		*/
static double Q1[N], Q2[N], Q3[N], Q4[N], QA[N], QB[N], QD[N];
static double QF1, QF2, QF3, QSL, QSR;
static int QNE, QNY, QFlag;

/* these will be set by PRAXIS to point to it's arguments */
static int n;
double *x;
double (*fun)(double *, int);

/* these will be set by PRAXIS to the global control parameters */
static double h, macheps;

double random()	/* return random no between 0 and 1 */ {
   return (double)(rand()%(8192*2))/(double)(8192*2);
}

void sort() {
/*  Sorts the elements of D and corresponding  */
/*  columns of V into descending order;        */

   int k, i, j;
   double s;

   for (i=0; i<n-1; i++) {
       k = i; s = d[i];
       for (j=i+1; j<n; j++) {
/*     Vom Rest den maximalen d(k) finden */
           if (d[j] > s) {
			   k = j; s = d[j];
		   }
       }
       if (k > i) {
/*	   Zeilen D(m) & V(*,m) umtauschen mit m = i und k	*/
		   d[k] = d[i]; d[i] = s;
		   for (j=0; j<n; j++) {
			   s = v[j][i]; v[j][i] = v[j][k]; v[j][k] = s;
		   }
       }
   }
}

void vecprint(char *s, double px[], int n) {
   int i;

   printf("%s\n", s);
   for (i=0; i<n; i++)
       printf("%20.10e ", px[i]);
   printf("\n");
}

void print()		/* print a line of traces */ {
   printf("\n");
   printf("... chi square reduced to ... %20.10e\n", fx);
   printf("... after %u function calls ...\n", nf);
   printf("... including %u linear searches ...\n", nl);
   vecprint("... current values of x ...", x, n);
}

void matprint(char *s, double v[][N], int n) {
   int k, i;

   printf("%s\n", s);
   for (k=0; k<n; k++) {
       for (i=0; i<n; i++) {
           printf("%20.10e ", v[k][i]);
       }
       printf("\n");
   }
}

void PQuad(double l1) {
	int i;  double Temp;
	Temp = l1 / QSR;
    for (i=0; i<n; i++)
    	tflin[i] = ((QD[i]*Temp + QA[i])*Temp + QB[i])*Temp + x[i];
}

double flin(double l, int j) {
   int i;
   if (j != -1) {	/* linear search */
      for (i=0; i<n; i++)
          tflin[i] = x[i] + l*v[i][j];
   }
   else {			/* search along parabolic space curve */
	   PQuad(l);
   }
   nf++;
   return (*fun)(tflin, n);

}

void minlin(int j, int nits, double *d2e, double *x1e, double f1, double fk) {
   int k, i, dz;  char *sd;
   double x1, x2, xm, f0, f2, fm, d1, d2, t2, s, sf1, sx1, d2s;

   x1 = *x1e;  d2 = *d2e;
   /*	Increment one-dimensional search counter	*/
   nl++;
   sf1 = f1;  sx1 = x1;
   k = 0;  xm = 0.0;
   fm = f0 = fx;
   dz = (d2 < macheps);
   /*	find step size							*/
   s = 0;
   for (i=0; i<n; i++) s += x[i]*x[i];
   s = sqrt(s);
   t2 = m4*sqrt(fabs(fx)/(dz? dmin: d2) + s*ldt) + m2*ldt;
   s = s*m4 + tol;
   if (dz && t2 > s) t2 = s;
   if (t2 < small) t2 = small;
   if (t2 > 0.01*h) t2 = 0.01 * h;

   if (!fk || fabs(x1) < t2) {
      x1 = (x1 >= 0 ? t2 : -t2);
      f1 = flin(x1, j);
   }
   if (f1 <= fm) {
      xm = x1;  fm = f1;
   }
   if (dz) {
   /*	Evaluate FLIN at another point and
        estimate the second derivative;			*/
      x2 = (f0 < f1)? -x1: 2*x1;
      f2 = flin(x2, j);
      if (f2 <= fm) {
         xm = x2;  fm = f2;
      }
   }

l0_next:
   if (dz) {
      d2 = (x2*(f1-f0) - x1*(f2-f0))/(x1*x2*(x1-x2));
   }
   /*	Estimate first derivative at 0			*/
   d1 = (f1-f0)/x1 - x1*d2;
   dz = 1;
   /*	Predict minimum							*/
   if (d2 <= small)
      x2 = (d1 < 0) ? h : -h;
   else {
      x2 = - 0.5 * d1 / d2;
      if (fabs(x2) > h)
		  x2 = (x2 > 0) ? h : -h;
   }
   /*	Evaluate F at the predicted minimum		*/
l1_try:
   f2 = flin(x2, j);
   if ((k < nits) && (f2 > f0)) {
   /*	No success so try again					*/
      k++;
      if ((f0 < f1) && (x1 * x2 > 0.0))
         goto l0_next;
      x2 *= 0.5;
      goto l1_try;
   }
   if (f2 <= fm) {
       xm = x2;  fm = f2;
   }
   /*	Get new estimate of second derivative	*/
   if (fabs(xm*(xm-x1)) > small) {
	  d2s = d2;
      d2 = (xm*(f1-f0) - x1*(fm-f0))/(x1*xm*(x1-xm));
      if (k == 0 && d2 > small) {
    	  s = d2s / d2;
    	  if (s > 1.04 || s < 0.92) {
    		  k++;  dz = 0;  goto l0_next;
    	  }
      }
   }
   else {
      if (k > 0) d2 = 0;
   }
   if (d2 < small) d2 = small;
   *x1e = xm;  fx = fm;  *d2e = d2;
   if (sf1 < fx) {
      fx = sf1;  *x1e = sx1;
   }
   for (i=0; i<n; i++) x[i] += (*x1e)*v[i][j];
}

void MinQuad(double *x1e, double f1, double f0, double ht) {
	int k;
	double x1, x2, xm, f2, fm, d1, d2s, t2;
	k = 0;  xm = 0;  fm = f0;
/*  Increment one-dimensional search counter  */
	nl++;

/*  Evaluate FLIN at another point and
    estimate the second derivative;     */
	x1 = *x1e;  x2 = -x1;
	f2 = flin(x2, -1);
	if (f2 <= fm) {xm = x2;  fm = f2;}
	d2s = (x2*(f1 - f0) - x1*(f2 - f0)) / (x1*x2*(x1 - x2));

/*  Estimate first derivative at 0		*/
	d1 = (f1 - f0)/x1 - x1*d2s;

/*  Predict minimum	 */
	if (d2s > small) {
		t2 = -0.5*d1/d2s;  if (t2 > h) t2 = ht;
		if (t2 < 0.) t2 = 0.4*x2;
		if (t2 < 0.14*x2) t2 = 0.28*t2;
		x2 = t2;
	}
	else x2 = ht;

/*  Evaluate F at the predicted minimum		*/
l1_next:
	f2 = flin(x2, -1);
	if (k < 2 && f2 > f0) {
/*      No success so try again	*/
		k++;  x2 *= 0.5;  goto l1_next;
	}

	if (f2 > f0) {x1 = xm;  fx = fm;}
	else {x1 = x2;  fx = f2;}
	*x1e = x1;  PQuad(x1);
	for (i=0; i<n; i++)  x[i] = tflin[i];
}

void Parabel_4P() {
	int i;
	double r2, r3, r4, r23, r34, s, u, t1, t2;
    double qk, qs, qz;
    r2 = r3 = r4 = r23 = r34 = 0.;
    for (i=0; i<n; i++) {
    	s = Q2[i] - Q1[i];  r2 += s*s;
    	s = Q3[i] - Q1[i];  r3 += s*s;
    	s = Q4[i] - Q1[i];  r4 += s*s;
    	s = Q3[i] - Q2[i];  r23 += s*s;
    	s = Q4[i] - Q3[i];  r34 += s*s;
    }
    r2 = sqrt(r2);  r3 = sqrt(r3);  r4 = sqrt(r4);  QSR = r2;
    r23 = sqrt(r23);  r34 = sqrt(r34);  QSL = r23 + r34;

/*  -- Prfen, ob konvex:  */
    s = 0.;
    for (i=0; i<n; i++) {
    	u = (Q3[i] - Q2[i]) / r23;
        s += (u - (Q2[i] - Q1[i])/r2) * ((Q4[i] - Q3[i])/r34 - u);
    }
	if (s <= macheps) { QFlag = 1;  return; }

/*  -- k berechnen, Zhler:  */
    s = 0.;
    for (i=0; i<n; i++) s += (Q2[i] - Q1[i]) * (Q4[i] - Q1[i]);
    qk = sqrt(r2*r2*r4*r4 - s*s);

/*  -- s berechnen, Zhler:  */
    s = 0.;
    for (i=0; i<n; i++) s += (Q3[i] - Q1[i]) * (Q4[i] - Q1[i]);
	qs = sqrt(r3*r3*r4*r4 - s*s);

/*  -- k und s, gemeinsamer Nenner:  */
    s = 0.;
    for (i=0; i<n; i++) s += (Q2[i] - Q1[i]) * (Q3[i] - Q1[i]);
    qz = sqrt(r2*r2*r3*r3 - s*s);

    qk /= qz;  qs /= qz;

/*  -- t1 berechnen:  */
    t1 = (sqrt(qs/qk * (qs - qk + 1.)) + qs) / (1. - qk);

    s = (t1 + 1)*t1;
    for (i=0; i<n; i++) {
    	QA[i] = (Q3[i] - Q1[i] + (Q2[i] - Q1[i]) * t1) / s;
        QB[i] = QA[i] - Q2[i] + Q1[i];
        QD[i] = 0.;
    }
    if (n > 2) {
/*  -- Die Torsion (Drehung) bercksichtigen:  */
		t2 = qk*t1 + qs;  u = 1. / (t2*(t2 + 1.)*(t2 - t1));
	    for (i=0; i<n; i++) {
	        s = Q4[i] - ((QA[i] * t2 + QB[i]) * t2 + x[i]);  /* Abweichung  */
	        s = 0.44 * s * u;   QD[i] = s;
	        QA[i] += s * (1. - t1);  QB[i] -= s * t1;
	    }
    }
}

void RotateV(double L) {
/*  Die V-Vektoren nach dem Parabel-Schritt hinterher drehen -
	von V(*,N) zu neuer Tangente. L - Argument der Parabel-Funktion. */
	int i, j;
	double tx, ty, u0x, u0y, u1x, u1y, Temp, s;
    double qx[N], qy[N];
    Temp = L / QSR;  tx = 0.;
    for (i=0; i<n; i++) {
    	qx[i] = v[i][n-1];
/*      -- Ableitung der Parabel im Punkt L => qy(*):  */
    	s = (3.*QD[i]*Temp + 2.*QA[i])*Temp + QB[i];
    	qy[i] = s;  tx += s*qx[i];
    }
/*  -- QX und QY bilden die Drehebene. |QX| ist schon = 1 */
    s = 0.;
    for (i=0; i<n; i++) {
    	Temp = qy[i] - qx[i]*tx;
    	qy[i] = Temp;  s += Temp*Temp;
    }
    s = sqrt(s);  ty = s;          		/* QY ist jetzt quer zu QX */
    for (i=0; i<n; i++) qy[i] /= s;		/* und jetzt auch Laenge 1 */
    Temp = sqrt(tx*tx + ty*ty);  tx = tx/Temp;  ty = ty/Temp;

/*  -- QX ist jetzt der x-Ort, QY ist der y-Ort  */
/*  -- tx = Cos, ty = Sin.  Drehen:  */
    for (j=0; j<n; j++) {
    	u0x = 0;  u0y = 0;
    	for (i=0; i<n; i++) {
    		u0x = u0x + v[i][j]*qx[i];
    		u0y = u0y + v[i][j]*qy[i];
    	}
    	u1x = u0x*tx - u0y*ty;
    	u1y = u0x*ty + u0y*tx;
    	for (i=0; i<n; i++)
    		v[i][j] += qx[i]*(u1x - u0x) + qy[i]*(u1y - u0y);
    }
}

int Quad() {	/* look for a minimum along the curve Q4, Q3, Q2, X */
	int i;
	double s, u, ss, ql;

/*  Q1 ist der beste Minimum-Punkt vor den ersten n-1	*/
/*	linearen Minimisierungen,  X ist der Punkt danach.	*/
	s = u = 0.;
	for (i=0; i<n; i++) {
		ss = x[i] - Q1[i];    s += ss*ss;
		ss = x[i] - Q2[i];    u += ss*ss;
		Q1[i] = x[i];
	}
    if (QFlag == 0)
/*    fx > QF2 ist mglich, wenn ILLC und Random (Ridge Problem) im Spiel sind */
    	if (s < 0.5 * u && fx < QF2)
    		QFlag = 1;
    	else {
    		QFlag = -1;
    		return(0);
    	}
    else
    	if (QFlag == 2)
    		if (s > 0.2 * u) QFlag = 1;

	QSR = sqrt(u);  ss = sqrt(s);
	if (QFlag == 2) {
		Parabel_4P();  			/* Q4---Q3---Q2---Q1=X---> */
		u = 1;
	}
	if (QFlag == 1) {
		s = QSL / QSR;  u = s + 1;
		for (i=0; i<n; i++) {
			QA[i] = Q3[i]/(s*u) - Q2[i]/s + Q1[i]/u;
			QB[i] = QA[i] + Q1[i] - Q2[i];
			QD[i] = 0.;
		}
	}
	ql = -QSR;  QF1 = fx;
	s = (QSL + QSR) * 1.6;
	if (s > h) s = h;

	QNE++;
	MinQuad(&ql, QF2, QF1, s);

	s = 0.;
	if (ql > macheps)
		for (i=0; i<n; i++) {
			Q4[i] = Q3[i];  Q3[i] = Q2[i];  Q2[i] = Q1[i];
			u = x[i];  Q1[i] = u;
			u -= Q2[i];  s += u*u;
		}

	if (ql < 0.) QFlag = -1;
	else if (QFlag == 1) {
		if (s < 0.007*h*h) QFlag = -1; }
	else if (QFlag == 2)
		if (s < 0.1*ss) QFlag = -1;
	if (QFlag <= 0)  return(0);

/*  Die V-Vektoren nach dem Kurvenablauf hinterher drehen */
	RotateV(ql);

	QSL = QSR;  QSR = sqrt(s);
	QF2 = QF1;  QF1 = fx;
	QNY++;  QFlag = 2;
	return(1);
}

double Praxis(double (*_fun)(double *, int), double _x[N], double _h, int _n)
{
   /* init global extern variables and parameters */
   fun = _fun;  x = _x;  h = _h;  n = _n;

   illc = 0;
   macheps = EPSILON;
   small = macheps*macheps;  vsmall = small*small;
   large = 1.0/small;  vlarge = 1.0/vsmall;
   m2 = sqrt(macheps);  m4 = sqrt(m2);

   ldfac = (illc ? 0.1 : 0.01);
   QNY = QNE = 0;  QFlag = -1;
   nl = kt = 0;  nf = 1;  fx = (*fun)(x, n);
   t2 = small + fabs(TPRIM);
   tol = t2;
   dmin = small;

   if (h < 100*tol) h = 100*tol;
   ldt = h;
   for (i=0; i<n; i++) {
	   for (j=0; j<n; j++)  v[i][j] = (i == j ? 1.0 : 0.0);
	   z[i] = 0;
   }
   d[0] = 0.0;
   if (prin > 1) {
      printf("\n------------- enter function PRAXIS -----------\n");
      printf("... current parameter settings ...\n");
      printf("... scaling ... %20.10e\n", scbd);
      printf("...   tol   ... %20.10e\n", tol);
      printf("... maxstep ... %20.10e\n", h);
      printf("...   illc  ... %20u\n", illc);
      printf("...   ktm   ... %20u\n", ktm);
   }
   if (prin) print();

mloop:
   sf = d[0];
   s = d[0] = 0.0;

   /* minimize along first direction */
   minlin(0, 2, &d[0], &s, fx, 0);
   if (s <= 0.0)
      for (i=0; i < n; i++)
          v[i][0] = -v[i][0];
   if ((sf <= (0.9 * d[0])) || ((0.9 * sf) >= d[0]))
      for (i=1; i<n; i++)
          d[i] = 0.0;
   for (k=1; k<n; k++) {
       for (i=0; i<n; i++)
           y[i] = x[i];
       sf = fx;
       illc = illc || (kt > 0);
l1_next:
       kl = k;
       df = 0.;
       if (illc) {        /* random step to get off resolution valley */
          for (i=0; i<n; i++) {
			  s = random();
              s = (0.1*ldt + t2 * pow(10., kt)) * (s - 0.5);
              z[i] = s;
              for (j=0; j < n; j++)
                  x[j] += s * v[j][i];
  	      }
          fx = (*fun)(x, n);
          nf++;
       }
       /* minimize along non-conjugate directions */
       for (k2=k; k2<n; k2++) {
    	   if (QFlag >= 0 && k == 1 && k2 == n-1)
    		   if (Quad()) goto mloop;
           sl = fx;
           s = 0.0;
           minlin(k2, 2, &d[k2], &s, fx, 0);
           if (illc) {
	          double szk = s + z[k2];
              s = d[k2] * szk*szk;
	       }
           else
	          s = sl - fx;
           if (df < s) {
              df = s; kl = k2;
           }
       }
       if (!illc && (df < fabs(100.0 * macheps * fx))) {
          illc = 1;
          goto l1_next;
       }
       if ((k == 1) && (prin > 1))
          vecprint("\n... New Direction ...",d,n);
       /* minimize along conjugate directions */
       for (k2=0; k2<k; k2++) {
           s = 0.0;
           minlin(k2, 2, &d[k2], &s, fx, 0);
       }
       if (k == n-1) {
    	   for (i=0; i<n; i++) {
    		   Q2[i] = x[i];  Q3[i] = y[i];
    	   }
    	   QF2 = fx;  QF3 = sf;
       }
       f1 = fx;
       fx = sf;
       lds = 0.0;
       for (i=0; i<n; i++) {
           sl = x[i];
           x[i] = y[i];
           y[i] = sl - y[i];
           sl = y[i];
           lds = lds + sl*sl;
       }
       lds = sqrt(lds);
       if (lds > small) {
          /*  Throw away direction KL and minimize
              along the new "conjugate" direction  */
          for (i=kl-1; i>=k; i--) {
              for (j=0; j < n; j++)
                  v[j][i+1] = v[j][i];
                  d[i+1] = d[i];
              }
              d[k] = 0.0;
              for (i=0; i < n; i++)
                  v[i][k] = y[i] / lds;
              minlin(k, 4, &d[k], &lds, f1, 1);
              if (lds <= 0.0) {
                 lds = -lds;
                 for (i=0; i<n; i++)
                     v[i][k] = -v[i][k];
              }
       }
       ldt = ldfac * ldt;
       if (ldt < lds) ldt = lds;
       if (prin > 1) print();
       t2 = 0.0;
       for (i=0; i<n; i++)
           t2 += x[i]*x[i];
       t2 = m2 * sqrt(t2) + tol;
       if (ldt > (0.5 * t2)) kt = 0;
       else kt++;
       if (kt > ktm)
          goto fret;
   }
   dn = 0.0;
   for (i=0; i<n; i++) {
       d[i] = 1.0 / sqrt(d[i]);
       if (dn < d[i]) dn = d[i];
   }
   if (prin > 2)
      matprint("\n... New Matrix of Directions ...",v,n);
   for (j=0; j<n; j++) {
       s = d[j] / dn;
       for (i=0; i < n; i++) v[i][j] *= s;
   }
   if (scbd > 1.0) {       /* scale axis to reduce condition number */
      s = vlarge;
      for (i=0; i<n; i++) {
          sl = 0.0;
          for (j=0; j < n; j++)
              sl += v[i][j]*v[i][j];
          z[i] = sqrt(sl);
          if (z[i] < m4) z[i] = m4;
          if (s > z[i]) s = z[i];
      }
      for (i=0; i<n; i++) {
          sl = s / z[i];
          z[i] = 1.0 / sl;
          if (z[i] > scbd) {
             sl = 1.0 / scbd; z[i] = scbd;
          }
/* jd: Folgende Zeile fehlte, auch in der Pascal-Version.	*/
/*     Allerdings keine Auswirkung auf die Test-Resultate!?	*/
		  for (j=0; j < n; j++) v[i][j] *= sl;
      }
   }
   for (i=1; i<n; i++)
       for (j=0; j<=i-1; j++) {
           s = v[i][j];
           v[i][j] = v[j][i];
           v[j][i] = s;
       }
   minfit(n, macheps, vsmall, v, d);
   if (scbd > 1.0) {
      for (i=0; i<n; i++) {
          s = z[i];
          for (j=0; j<n; j++)
              v[i][j] *= s;
      }
      for (i=0; i<n; i++) {
          s = 0.0;
          for (j=0; j<n; j++)
              s += v[j][i]*v[j][i];
          s = sqrt(s);
          d[i] *= s;
          s = 1.0 / s;
          for (j=0; j<n; j++)
              v[j][i] *= s;
      }
   }
   for (i=0; i<n; i++) {
       if ((dn * d[i]) > large)
          d[i] = vsmall;
       else if ((dn * d[i]) < small)
          d[i] = vlarge;
       else
          d[i] = pow(dn * d[i],-2.0);
   }
   sort();               /* the new eigenvalues and eigenvectors */
   /*  try quadratic extrapolation in case    */
   /*  we are stuck in a curved valley        */
   if (d[n-2] > 3.3 * d[n-1]) {
	   QSL = QSR = 0.;  QF1 = fx;
	   /* Was hier getan wird:
	      Q1 = X;  QSL = |Q2 - Q3|;  QSR = |Q1 - Q2|; */
	   for (i=0; i<n; i++) {
		   s = Q2[i];  sl = x[i];  Q1[i] = sl;
		   sl -= s;  s -= Q3[i];  QSL += s*s;  QSR += sl*sl;
	   }
	   QSL = sqrt(QSL);  QSR = sqrt(QSR);
	   if (QSR > 0.025*h && QSL > 0.00025*QSR) QFlag = 0;
   }
   dmin = d[n-1];
   if (dmin < small) dmin = small;
   illc = (m2 * d[0]) > dmin;
   if ((prin > 2) && (scbd > 1.0))
      vecprint("\n... Scale Factors ...",z,n);
   if (prin > 2)
      vecprint("\n... Eigenvalues of A ...",d,n);
   if (prin > 2)
      matprint("\n... Eigenvectors of A ...",v,n);
   goto mloop; 	 /* back to main loop */

fret:
   if (prin > 0) {
         vecprint("\n... Final solution is ...", x, n);
         printf("\n... ChiSq reduced to %20.10e ...\n", fx);
	 printf("... after %20u function calls.\n", nf);
   }
   return(fx);
}
